/*-
 *  Copyright (c) 2001  Peter Pentchev
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "vpoppassd.h"
#include "vpp_err.h"
#include "vpp_flavor.h"
#include "vpp_util.h"

static char cvs_id[] __unused =
"$Ringlet: vpoppassd.c 2629 2008-10-09 09:58:45Z roam $";

typedef vpp_err_t	(*vpp_command)(const char *cmd, const char *args);

int			 verbose, quiet;

char			 vpp_verstr[64] = "vpoppassd";

const char		*vpp_uservar;
const char		*vpp_user;
const char		*vpp_passwd;
const char		*vpp_newuser;

static const char	*vpp_flav_in;
static const char	*vpp_flav_out;

static vpp_err_t	vpp_init(int, const char *[]);
static vpp_err_t	vpp_doit(void) __dead2;

static vpp_err_t	vpp_env_init(void);

static void		usage(void);

static vpp_err_t	vpp_cmd_newpass(const char *cmd, const char *args);
static vpp_err_t	vpp_cmd_quit(const char *cmd, const char *args);

#ifdef ENABLE_CUSER
static vpp_err_t	vpp_cmd_cuser(const char *cmd, const char *args);
static int		vpp_cuser_auth(const char *cuser, const char *nuser);
#endif /* ENABLE_CUSER */

#ifdef ENABLE_NEWUSER
static vpp_err_t	vpp_cmd_newuser(const char *cmd, const char *args);
static int		vpp_newuser_auth(const char *newuser, const char *nuser);
#endif /* ENABLE_NEWUSER */

static struct {
	const char	*name;
	vpp_command	 cmd;
} commands[] = {
	{"NEWPASS",	vpp_cmd_newpass},
	{"QUIT",	vpp_cmd_quit},
#ifdef ENABLE_CUSER
	{"CUSER",	vpp_cmd_cuser},
#endif /* ENABLE_CUSER */
#ifdef ENABLE_NEWUSER
	{"NEWUSER",	vpp_cmd_newuser},
#endif /* ENABLE_NEWUSER */
	{NULL,		NULL},
};

/*
 * Function:
 *         vpp_makeversion()                - generate version string
 * Inputs:
 *         none
 * Returns:
 *         vpp_err_t
 * Modifies:
 *         vpp_verstr
 */

static vpp_err_t
vpp_makeversion(void) {

	snprintf(vpp_verstr, sizeof(vpp_verstr), "vpoppassd v%d.%d"
#if VPP_VER_PRE
	    "-pre%d"
#endif /* VPP_VER_PRE */
#if VPP_VER_PATCH
	    "p%d"
#endif /* VPP_VER_PATCH */
	    ,
	    VPP_VER_MAJ, VPP_VER_MIN
#if VPP_VER_PRE
	    , VPP_VER_PRE
#endif /* VPP_VER_PRE */
#if VPP_VER_PATCH
	    , VPP_VER_PATCH
#endif /* VPP_VER_PATCH */
		);
	
	return (VPP_ERR_NONE);
}

/*
 * Function:
 * 	vpp_init				- general-purpose init function
 * Inputs:
 * 	argc, argv			- main() args for option processing
 * Returns:
 * 	vpp_err_t
 * 	CMDLINE
 * Modifies:
 * 	verbose, quiet
 */

static vpp_err_t
vpp_init(int argc, const char *argv[]) {
	vpp_err_t r;
	
	vpp_makeversion();
	
	/* Process environment variables */

	if (r = vpp_env_init(), r)
		return (r);

	/* Input flavor */
	if (vpp_flav_in != NULL)
		r = vpp_flavor_set(VPP_FLT_IN, vpp_flav_in, argc, argv);
	else
		r = vpp_flavor_probe(VPP_FLT_IN, argc, argv);
	if (r != VPP_ERR_NONE)
		return (r);

	/* Output flavor */
	if (vpp_flav_out != NULL)
		r = vpp_flavor_set(VPP_FLT_OUT, vpp_flav_out, argc, argv);
	else if (argc == 1)
		usage();
	else
		r = vpp_flavor_probe(VPP_FLT_OUT, argc, argv);
	if (r != VPP_ERR_NONE)
		return (r);

	return (VPP_ERR_NONE);
}

/*
 * Function:
 * 	usage			- startup help info
 * Inputs:
 * 	none
 * Returns:
 * 	nothing
 * Modifies:
 * 	nothing, writes to stdout and exits
 */

void
usage(void) {
	static const char *msg[] = {
	    "usage: vpoppassd [command [args..]]",
	    "",
	    "All configuration is done from environment variables.",
	    "At least the VPP_FLAVOR_IN and VPP_FLAVOR_OUT variables"
	    " should be set.",
	    NULL
	};
	unsigned i;
	
	for (i = 0; msg[i] != NULL; i++)
		fprintf(stderr, "%s\n", msg[i]);
	exit(VPP_ERR_CMDLINE);
}

/*
 * Function:
 * 	vpp_doit				- perform the actual work
 * Inputs:
 * 	argc, argv			- main() args
 * Returns:
 * 	vpp_err_t
 * 	nothing so far (FIXME)
 * Modifies:
 *	nothing by itself; writes args to stdout
 */

static vpp_err_t
vpp_doit(void) {
	char *line;
	size_t sz, cmdlen, argofs;
	unsigned i;
	vpp_err_t r;
	
	if (r = vpp_writeline(STDOUT_FILENO, 1, "+OK"), r)
		exit(vpp_prerror("writing to stdout", r));

	setvbuf(stdin, (char *) NULL, _IOLBF, 0);
	RDBG(("The username we have is %s\n", vpp_user));
	for (;;) {
		if (r = vpp_readline(STDIN_FILENO, &line, &sz),
		    r != VPP_ERR_NONE)
			exit(vpp_prerror("reading input", r));
		if (line == NULL)
			exit(VPP_ERR_NONE);

		while ((sz > 0) &&
		    ((line[sz - 1] == '\r') || (line[sz - 1] == '\n')))
			line[--sz] = '\0';
		RDBG(("Read cmd %s\n", line));

		for (cmdlen = 0; cmdlen < sz; cmdlen++)
			if (isspace(line[cmdlen]))
				break;
		if (cmdlen == 0) {
			/* empty command */
			free(line);
			continue;
		}

		for (argofs = cmdlen; argofs < sz; argofs++)
			if (!isspace(line[argofs]))
				break;

		RDBG(("Command %.*s, args %s\n",
		    (int) cmdlen, line, line + argofs));

		for (i = 0; commands[i].name != NULL; i++)
			if (!strncasecmp(commands[i].name, line, cmdlen) &&
			    commands[i].name[cmdlen] == '\0')
				break;
		if (commands[i].name == NULL) {
			if (r = vpp_writeline(STDOUT_FILENO, 1,
			    "-ERR Unknown command"), r)
				exit(vpp_prerror("writing to stdout", r));
		} else {
			line[cmdlen] = '\0';
			if (r = commands[i].cmd(line, line + argofs), r)
				exit(vpp_prerror("executing command", r));
			RDBG(("Processing of cmd %s, args %s finished\n",
			    line, line + argofs));
		}
		free(line);
	}
	/* NOTREACHED */
}

static vpp_err_t
vpp_env_init(void) {
	const char *v;

	if (v = getenv("VPP_USERVAR"), v != NULL)
		vpp_uservar = v;

	if (v = getenv("VPP_FLAVOR_IN"), v != NULL)
		vpp_flav_in = v;
	if (v = getenv("VPP_FLAVOR_OUT"), v != NULL)
		vpp_flav_out = v;

	return (VPP_ERR_NONE);
}

static vpp_err_t
vpp_cmd_newpass(const char *cmd __unused, const char *args) {

	if (args[0] == '\0')
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR Need a password"));

	/* FIXME: I really have to find a way to free the old value.. */
	vpp_passwd = strdup(args);

	if (vpp_newuser == NULL) {
		RDBG(("Attempting to change %s's password to %s..\n",
		    vpp_user, vpp_passwd));
		return (vpp_flavor_change(args));
	} else {
		RDBG(("Attempting to create a new user %s, password %s..\n",
		    vpp_newuser, vpp_passwd));
		return (vpp_flavor_newuser(args));
	}
}

#ifdef ENABLE_CUSER

static vpp_err_t
vpp_cmd_cuser(const char *cmd __unused, const char *args) {
	vpp_err_t r;

	if (args[0] == '\0')
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR Need a username"));

	if (!vpp_cuser_auth(vpp_user, args))
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR You are not allowed to issue a CUSER command"));

	/* FIXME: I really have to find a way to free the old value.. */
	vpp_user = strdup(args);

	if (r = vpp_writeline(STDOUT_FILENO, 1,
	    "+OK Username changed to %s", vpp_user), r)
		return (r);
	return (VPP_ERR_NONE);
}

static int
vpp_cuser_auth(const char *cuser, const char *nuser) {
	size_t len;
	
	/* FIXME: should this really be allowed? */
	if (!strcmp(cuser, nuser))
		return (1);

	/* FIXME: a hardcoded operator username for the present.. */
	len = strlen(CUSER_INVOKE);
	if (strncmp(cuser, CUSER_INVOKE, len) != 0)
		return (0);
	if ((cuser[len] != '\0') &&
	    (cuser[len] != '@') && (cuser[len] != '%'))
		return (0);

	return (1);
}

#endif /* ENABLE_CUSER */

#ifdef ENABLE_NEWUSER

static vpp_err_t
vpp_cmd_newuser(const char *cmd __unused, const char *args) {
	vpp_err_t r;

	if (args[0] == '\0')
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR Need a username"));

	if (!vpp_newuser_auth(vpp_user, args))
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR You are not allowed to issue a NEWUSER command"));

	/* FIXME: I really have to find a way to free the old value.. */
	vpp_newuser = strdup(args);

	if (r = vpp_writeline(STDOUT_FILENO, 1,
	    "+OK About to create new user %s", vpp_newuser), r)
		return (r);
	return (VPP_ERR_NONE);
}

static int
vpp_newuser_auth(const char *newuser, const char *nuser __unused) {
	size_t len;
	
	/* FIXME: a hardcoded operator username for the present.. */
	len = strlen(NEWUSER_INVOKE);
	if (strncmp(newuser, NEWUSER_INVOKE, len) != 0)
		return (0);
	if ((newuser[len] != '\0') &&
	    (newuser[len] != '@') && (newuser[len] != '%'))
		return (0);

	/* FIXME: should we not check for the same mail domain? */

	return (1);
}

#endif /* ENABLE_NEWUSER */

static vpp_err_t
vpp_cmd_quit(const char *cmd __unused, const char *args __unused) {

	exit(vpp_writeline(STDOUT_FILENO, 1, "+OK Bye!"));
}

/*
 *   M A I N   F U N C T I O N
 */

int
main(int argc, const char *argv[]) {
	vpp_err_t r;
	
	if (r = vpp_init(argc, argv), r)
		return (vpp_prerror("init", r));
	
	vpp_doit();
	/* NOTREACHED */
}
