/*-
 *  Copyright (c) 2001, 2008  Peter Pentchev
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/wait.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "vpoppassd.h"
#include "vpp_flavor.h"
#include "vpp_fl_vpasswd.h"
#include "vpp_util.h"

#ifdef EBUG
#include <stdio.h>
#endif

static char	cvs_id[] __unused =
"$Ringlet: vpp_fl_vpasswd.c 2658 2008-10-09 13:56:30Z roam $";

static const char	*vpasswd_prog_def[] = {
	VPASSWD_PATH
};
#define VPASSWD_PROG_DEF	(sizeof(vpasswd_prog_def)/sizeof(vpasswd_prog_def[0]))

static const char	*vpasswd_args_def[] = {
	"%U",
	"%P",
};
#define VPASSWD_ARGS_DEF	(sizeof(vpasswd_args_def)/sizeof(vpasswd_args_def[0]))

static const char	*vadduser_prog_def[] = {
	VADDUSER_PATH
};
#define VADDUSER_PROG_DEF	(sizeof(vadduser_prog_def)/sizeof(vadduser_prog_def[0]))

static const char	*vadduser_args_def[] = {
	"%U",
	"%P",
};
#define VADDUSER_ARGS_DEF	(sizeof(vadduser_args_def)/sizeof(vadduser_args_def[0]))

static char		**vpasswd_prog;
static char		**vadduser_prog;

static vpp_err_t	vpasswd_probe(int argc, const char *argv[]);
static vpp_err_t	vpasswd_init(int argc, const char *argv[]);
static vpp_err_t	vpasswd_change(const char *pass);
static vpp_err_t	vpasswd_newuser(const char *pass);

static vpp_err_t
vpasswd_init_prog(char ***pprog, int argc, const char *argv[],
		const char *envprog, const char *envargs,
		unsigned defprogcnt, const char *defprog[],
		unsigned defargscnt, const char *defargs[]);

struct vpp_flavor_hnd	vpp_flh_vpasswd = {
	VPP_FLT_OUT,
	"vpasswd",
	VPP_FL_VPASSWD,
	vpasswd_probe,
	vpasswd_init,
	vpasswd_change,
	vpasswd_newuser
};

static vpp_err_t
vpasswd_probe(int argc, const char *argv[]) {

	if (argc == 1)
		return (VPP_ERR_FL_NOMATCH);
	if (strcmp(argv[1], "pw"))
		return (VPP_ERR_FL_NOMATCH);

	return (VPP_ERR_NONE);
}

static vpp_err_t
vpasswd_init(int argc, const char *argv[]) {
	vpp_err_t r;

	RDBG(("Building vpasswd_prog..\n"));
	if (r = vpasswd_init_prog(&vpasswd_prog, argc, argv,
	    getenv("VPASSWD_PROG"), getenv("VPASSWD_ARGS"),
	    VPASSWD_PROG_DEF, vpasswd_prog_def,
	    VPASSWD_ARGS_DEF, vpasswd_args_def), r)
		return (r);
	
	RDBG(("Building vadduser_prog..\n"));
	if (r = vpasswd_init_prog(&vadduser_prog, 0, NULL,
	    getenv("VADDUSER_PROG"), getenv("VADDUSER_ARGS"),
	    VADDUSER_PROG_DEF, vadduser_prog_def,
	    VADDUSER_ARGS_DEF, vadduser_args_def), r)
		return (r);

	return (VPP_ERR_NONE);
}

static vpp_err_t
vpasswd_init_prog(char ***pprog, int argc, const char *argv[],
		const char *envprog, const char *envargs,
		unsigned defprogcnt, const char *defprog[],
		unsigned defargscnt, const char *defargs[]) {
	char **v;
	unsigned c, i;
	vpp_err_t r;

	c = 0;
	v = NULL;

	if (r = vpp_argv_start(&c, &v), r)
		return (r);

	do {
		/* fake loop so I can break out */
		if (argc > 1) {
			if (r = vpp_argv_add(&c, &v, argc - 1, argv + 1), r)
				break;
		} else {
			if (envprog != NULL) {
				/* break it down.. */
				return (VPP_ERR_INT);
			} else {
				if (r = vpp_argv_add(&c, &v,
				    defprogcnt, defprog), r)
					break;
			}
			if (envargs != NULL) {
				/* break it down.. */
				return (VPP_ERR_INT);
			} else {
				if (r = vpp_argv_add(&c, &v,
				    defargscnt, defargs), r)
					break;
			}
		}

		for (i = 0; i < c; i++)
			RDBG(("v[%d] = %s\n", i, v[i]));
	} while (0);

	*pprog = v;
	return (VPP_ERR_NONE);
}

static vpp_err_t
vpasswd_change(const char *pass __unused) {
	int pipes[2];
	char *output, *noutput, *line;
	size_t nsz, sz;
	unsigned i;
	pid_t pid;
	int status;
	vpp_err_t r;

	/* Spawn pw(8).. */
	if (r = vpp_argv_subst(vpasswd_prog), r)
		return (r);
	if (r = vpp_spawn(pipes, vpasswd_prog[0], vpasswd_prog), r)
		return (r);
	if (pid = waitpid(-1, &status, WNOHANG), pid == -1)
		return (VPP_ERR_WAIT);
	RDBG(("waitpid: pid=%d, status=%X\n", pid, status));
	/* Nothing to write. */

	/* Gather output */
	output = NULL;
	sz = 0;
	RDBG(("gathering output..\n"));
	while (r = vpp_readline(pipes[0], &line, &nsz), r == VPP_ERR_NONE) {
		if (line == NULL)
			break;

		for (i = 0; i < nsz; i++)
			if ((line[i] == '\r') || (line[i] == '\n'))
				line[i] = '.';
		RDBG(("Read %.*s\n", (int) sz, line));
		if (noutput = realloc(output, sz + nsz), noutput == NULL) {
			free(output);
			return (VPP_ERR_NOMEM);
		}
		memcpy(noutput + sz, line, nsz);
		output = noutput;
		sz += nsz;
	}
	RDBG(("finished gathering output, output=%p (%.*s), len=%d\n",
	    output, (int) sz, output, sz));

	if (pid == 0) {
		if (pid = waitpid(pid, &status, 0), pid == -1) {
			RDBG(("waitpid2: pid=%d, status=%X\n", pid, status));
			free(output);
			return (VPP_ERR_WAIT);
		}
	}
	RDBG(("test: status=%X, WIFEXITED=%d, WEXITSTATUS=%X\n",
	    status, WIFEXITED(status), WEXITSTATUS(status)));
	if (WIFEXITED(status) && (WEXITSTATUS(status) == 0))
		r = vpp_write(STDOUT_FILENO, "+OK ", 4);
	else
		r = vpp_write(STDOUT_FILENO, "-ERR ", 5);
	if (r != VPP_ERR_NONE)
		return (r);
	if (output != NULL) {
		r = vpp_write(STDOUT_FILENO, output, sz);
		free(output);
		output = NULL;
		if (r != VPP_ERR_NONE)
			return (r);
	}
	return (vpp_write(STDOUT_FILENO, "\r\n", 2));
}

static vpp_err_t
vpasswd_newuser(const char *pass __unused) {
	int pipes[2];
	char *output, *noutput, *line;
	size_t nsz, sz;
	unsigned i;
	pid_t pid;
	int status;
	vpp_err_t r;

	if (vpp_newuser == NULL)
		return (VPP_ERR_INT);
	vpp_user = vpp_newuser;

	/* Spawn pw(8).. */
	if (r = vpp_argv_subst(vadduser_prog), r)
		return (r);
	if (r = vpp_spawn(pipes, vadduser_prog[0], vadduser_prog), r)
		return (r);
	if (pid = waitpid(-1, &status, WNOHANG), pid == -1)
		return (VPP_ERR_WAIT);
	RDBG(("waitpid: pid=%d, status=%X\n", pid, status));
	/* Nothing to write. */

	/* Gather output */
	output = NULL;
	sz = 0;
	RDBG(("gathering output..\n"));
	while (r = vpp_readline(pipes[0], &line, &nsz), r == VPP_ERR_NONE) {
		if (line == NULL)
			break;

		for (i = 0; i < nsz; i++)
			if ((line[i] == '\r') || (line[i] == '\n'))
				line[i] = '.';
		RDBG(("Read %.*s\n", (int) sz, line));
		if (noutput = realloc(output, sz + nsz), noutput == NULL) {
			free(output);
			return (VPP_ERR_NOMEM);
		}
		memcpy(noutput + sz, line, nsz);
		output = noutput;
		sz += nsz;
	}
	RDBG(("finished gathering output, output=%p (%.*s), len=%d\n",
	    output, (int) sz, output, sz));

	if (pid == 0) {
		if (pid = waitpid(pid, &status, 0), pid == -1) {
			RDBG(("waitpid2: pid=%d, status=%X\n", pid, status));
			free(output);
			return (VPP_ERR_WAIT);
		}
	}
	RDBG(("test: status=%X, WIFEXITED=%d, WEXITSTATUS=%X\n",
	    status, WIFEXITED(status), WEXITSTATUS(status)));
	if (WIFEXITED(status) && (WEXITSTATUS(status) == 0))
		r = vpp_write(STDOUT_FILENO, "+OK ", 4);
	else
		r = vpp_write(STDOUT_FILENO, "-ERR ", 5);
	if (r != VPP_ERR_NONE)
		return (r);
	if (output != NULL) {
		r = vpp_write(STDOUT_FILENO, output, sz);
		free(output);
		output = NULL;
		if (r != VPP_ERR_NONE)
			return (r);
	}
	return (vpp_write(STDOUT_FILENO, "\r\n", 2));
}
