/*-
 *  Copyright (c) 2001  Peter Pentchev
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "vpoppassd.h"
#include "vpp_flavor.h"
#include "vpp_util.h"

#include "vpp_fl_bsdpw.h"
#include "vpp_fl_checkpwd.h"
#include "vpp_fl_vchkpw.h"
#include "vpp_fl_vpasswd.h"

#ifdef EBUG
#include <stdio.h> /* FIXME: debug only? */
#endif

static char	cvs_id[] __unused =
"$Ringlet: vpp_flavor.c 2629 2008-10-09 09:58:45Z roam $";

static struct vpp_flavor_hnd	*flav[] = {
	/* input */
	&vpp_flh_vchkpw,
	&vpp_flh_checkpwd,
	/* output */
	&vpp_flh_bsdpw,
	&vpp_flh_vpasswd,
	NULL
};

static struct vpp_flavor_hnd	*cflavor[VPP_FLT_LAST];

static vpp_err_t	vpp_flavor_set_byhnd(struct vpp_flavor_hnd *h,
		int argc, const char *argv[]);

/*
 * Function:
 * 	vpp_flavor_set			- set a specified flavor
 * Inputs:
 * 	flavor				- the flavor to set
 * Returns:
 * 	vpp_err_t
 * 	nothing so far (FIXME)
 * Modifies:
 * 	nothing so far (FIXME)
 */

vpp_err_t
vpp_flavor_set(vpp_flavor_type t, const char *flavor, int argc, const char *argv[]) {
	unsigned i;

	for (i = 0; flav[i] != NULL; i++) {
		if (flav[i]->name == NULL)
			return (VPP_ERR_INT);
		if ((flav[i]->type == t) && !strcmp(flav[i]->name, flavor))
			break;
	}
	if (flav[i] == NULL)
		return (VPP_ERR_FL_UNKNOWN);

	return (vpp_flavor_set_byhnd(flav[i], argc, argv));
}

/*
 * Function:
 * 	vpp_flavor_setdefault		- set a default flavor based on args
 * Inputs:
 * 	argc, argv			- main() args
 * Returns:
 * 	vpp_err_t
 * 	FL_UNKNOWN			- could not determine the flavor
 * 	error code from the various flavor probe routines
 * Modifies:
 * 	sets cflavor to the detected flavor type
 * 	calls the various flavor probe routines
 */

vpp_err_t
vpp_flavor_probe(vpp_flavor_type t, int argc, const char *argv[]) {
	unsigned i;
	vpp_err_t r;

	for (i = 0; flav[i] != NULL; i++)
		if ((flav[i]->type == t) && (flav[i]->probe != NULL)) {
			r = flav[i]->probe(argc, argv);
			if (r == VPP_ERR_NONE)
				break;
			else if (r != VPP_ERR_FL_NOMATCH)
				return (r);
		}
	if (flav[i] == NULL)
		return (VPP_ERR_FL_UNKNOWN);

	return (vpp_flavor_set_byhnd(flav[i], argc, argv));
}

static vpp_err_t
vpp_flavor_set_byhnd(struct vpp_flavor_hnd *h, int argc, const char *argv[]) {
	const char *val;
	vpp_err_t r;

	if (h->init == NULL)
		return (VPP_ERR_INT);
	if (r = h->init(argc, argv), r)
		return (r);

	/* Do whatever else is necessary.. */
	if (h->type == VPP_FLT_IN) {
		/* grab the username */
		if (vpp_uservar == NULL)
			return (VPP_ERR_INT);
		if (val = getenv(vpp_uservar), val == NULL)
			return (VPP_ERR_NOUSER);
		vpp_user = val;
	} else if (h->type == VPP_FLT_OUT) {
		/* nothing to do here.. I think. */
	}
	
	cflavor[h->type] = h;
	return (VPP_ERR_NONE);
}

vpp_err_t
vpp_flavor_change(const char *pass) {

	/* cflavor should have been set by vpp_init() long ago.. */
	if (cflavor[VPP_FLT_OUT] == NULL)
		return (VPP_ERR_INT);
	if (cflavor[VPP_FLT_OUT]->change == NULL)
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR Unsupported operation for the selected flavor"));
	
	return (cflavor[VPP_FLT_OUT]->change(pass));
}

vpp_err_t
vpp_flavor_newuser(const char *pass) {

	/* cflavor should have been set by vpp_init() long ago.. */
	if (cflavor[VPP_FLT_OUT] == NULL)
		return (VPP_ERR_INT);
	if (cflavor[VPP_FLT_OUT]->newuser == NULL)
		return (vpp_writeline(STDOUT_FILENO, 1,
		    "-ERR Unsupported operation for the selected flavor"));

	return (cflavor[VPP_FLT_OUT]->newuser(pass));
}
