/*-
 *  Copyright (c) 2001, 2008  Peter Pentchev
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#include <sys/types.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "vpoppassd.h"
#include "vpp_err.h"
#include "vpp_util.h"

static char	cvs_id[] __unused =
"$Ringlet: vpp_util.c 2637 2008-10-09 11:07:54Z roam $";

#define VPP_BUFSIZ	2048

vpp_err_t
vpp_argv_start(unsigned *pargc, char ***pargv) {
	unsigned c;
	char **v;

	c = 0;
	if (v = calloc(c + 1, sizeof(*v)), v == NULL)
		return (VPP_ERR_NOMEM);
	v[c] = NULL;

	*pargc = c;
	*pargv = v;
	return (VPP_ERR_NONE);
}

vpp_err_t
vpp_argv_add(unsigned *pargc, char ***pargv, unsigned count,
		const char *args[]) {
	unsigned c, i, j;
	char **v;
	char *val;
	vpp_err_t r;

	r = VPP_ERR_NONE;

	c = *pargc;
	if (v = realloc(*pargv, (c + count + 1) * sizeof(*v)), v == NULL)
		return (VPP_ERR_NOMEM);
	memset(v + c, 0, (count + 1) * sizeof(*v));
	/*
	 * Grumble.  Since we're using realloc(), we *must* let
	 * the caller know that the old pointer is invalid.
	 * This means modifying val/res params even in case of failure,
	 * but the alternative - not using realloc() - is even uglier.
	 */
	*pargv = v;
	
	for (i = 0; i < count; i++) {
		if (val = strdup(args[i]), val == NULL) {
			r = VPP_ERR_NOMEM;
			break;
		}
		/* Once again modifying even in case of failure.. */
		v[c + i] = val;
	}

	if (r != VPP_ERR_NONE) {
		/* And yet once again modifying even in case of failure.. */
		for (j = 0; j < i; j++) {
			free(v[c + j]);
			v[c + j] = NULL;
		}
		return (r);
	}

	*pargc = c + count;
	return (VPP_ERR_NONE);
}

/*
 * Function:
 * 	vpp_readline			- non-stdio-based line-by-line read
 * Inputs:
 * 	fd				- file descriptor to read from
 * 	pbuf				- val/res pointer for the line read
 * 	psz				- val/res pointer for the line length
 * Returns:
 * 	vpp_err_t
 * 	NOMEM, READ
 * Modifies:
 * 	allocates and stores into *pbuf, *psz
 */

vpp_err_t
vpp_readline(int fd, char **pbuf, size_t *psz) {
	static char *buf = NULL;
	static size_t ofs = 0, all = 0;
	static int lasttry = 0;
	char *line, *nl;
	size_t len;
	ssize_t n;

	buf = line = NULL;
	all = ofs = 0;

	for (;;) {
		/* First check for an inherited newline */
		if (lasttry ||
		    ((ofs > 0) && (nl = memchr(buf, '\n', ofs), nl != NULL))) {
			if (lasttry) {
				if (ofs == 0) {
					*pbuf = NULL;
					*psz = 0;
					return (VPP_ERR_NONE);
				}
				nl = buf + ofs;
				len = ofs;
			} else {
				/* Found a newline!  Go for it! */
				len = nl - buf + 1;
			}

			if (line = calloc(1, len + 1), line == NULL)
				return (VPP_ERR_NOMEM);
			memmove(line, buf, len);
			memmove(buf, buf + len, ofs - len);
			ofs -= len;

			*pbuf = line;
			*psz = len;
			return (VPP_ERR_NONE);
		}

		/* Need to allocate? */
		if (ofs == all) {
			/* Allocate the new chunk */
			line = realloc(buf, all + VPP_BUFSIZ);
			if (line == NULL)
				return (VPP_ERR_NOMEM);
			buf = line;
			all += VPP_BUFSIZ;
		}

		/* Read.. */
		if (n = read(fd, buf + ofs, all - ofs), n == -1) {
			free(buf);
			buf = NULL;
			return (VPP_ERR_READ);
		} else if (n == 0) {
			lasttry = 1;
		}
		ofs += n;
	}
	/* NOTREACHED */
}

vpp_err_t
vpp_spawn(int *ppipe, const char *prog, char *const argv[]) {
#ifdef EBUG
	unsigned i;
#endif
	int pipedown[2], pipeup[2], fdown, fup;
	pid_t pid;
	vpp_err_t r;

#ifdef EBUG
	RDBG(("spawn %s\n", prog));
	for (i = 0; argv[i] != NULL; i++)
		RDBG(("%s\n", argv[i]));
#endif

	fdown = fup = 0;

	do {
		r = VPP_ERR_PIPE;
		if (pipe(pipedown) == -1)
			break;
		fdown = 1;
		if (pipe(pipeup) == -1)
			break;
		fup = 1;
		
		if (pid = fork(), pid == -1) {
			r = VPP_ERR_FORK;
			break;
		}

		if (pid == 0) {
			/* Child */
			close(pipeup[0]);
			close(pipedown[1]);

			close(0);
			close(1);
			close(2);
			dup2(pipedown[0], 0);
			dup2(pipeup[1], 1);
			dup2(pipeup[1], 2);

			execvp(prog, argv);
			/* We should, like, never get here.. */
			exit(vpp_prerror(prog, VPP_ERR_EXEC));
		}

		close(pipeup[1]);
		close(pipedown[0]);
		r = VPP_ERR_NONE;
	} while (0);

	if (r != VPP_ERR_NONE) {
		if (fdown) {
			close(pipedown[0]);
			close(pipedown[1]);
		}
		if (fup) {
			close(pipeup[0]);
			close(pipeup[1]);
		}
		return (r);
	}

	ppipe[0] = pipeup[0];
	ppipe[1] = pipedown[1];
	return (VPP_ERR_NONE);
}

vpp_err_t
vpp_argv_subst(char *argv[]) {
	unsigned i;
	char *newval;

	for (i = 0; argv[i] != NULL; i++)
		if (argv[i][0] == '%') {
			switch (argv[i][1]) {
				case '%':
					newval = strdup(argv[i] + 1);
					break;

				case 'U':
					newval = strdup(vpp_user);
					break;

				case 'P':
					newval = strdup(vpp_passwd);
					break;

				default:
					/* FIXME: allocate an error code.. */
					return (VPP_ERR_INT);
			}

			if (newval == NULL)
				return (VPP_ERR_NOMEM);
			free(argv[i]);
			argv[i] = newval;
		}

	return (VPP_ERR_NONE);
}

vpp_err_t
vpp_write(int fd, const char *line, size_t len) {
	size_t written;
	ssize_t n;

	for (written = n = 0; written < len; written += n) {
		if (n = write(fd, line + written, len - written),
		    n == -1)
			return (VPP_ERR_WRITE);
		if (n == 0)
			return (VPP_ERR_NONE);
	}
	return (VPP_ERR_NONE);
}

vpp_err_t
vpp_writeline(int fd, int crlf, const char *fmt, ...) {
	va_list v;
	char *line;
	int n;
	vpp_err_t r;

	r = VPP_ERR_NONE;

	va_start(v, fmt);
	n = vasprintf(&line, fmt, v);
	va_end(v);
	if (n == -1 || line == NULL)
		return (VPP_ERR_NOMEM);

	if (r = vpp_write(fd, line, strlen(line)), r != VPP_ERR_NONE) {
		free(line);
		return (r);
	}
		
	if (crlf)
		r = vpp_write(fd, "\r\n", 2);
	else
		r = vpp_write(fd, "\n", 1);
	free(line);
	return (r);
}
